#Region "Disclaimer"
' By Teo Lachev "Applied Microsoft Analysis Services 2005"
'============================================================================
'   File:       StoredProcedures.vb
'   Summary:    SSAS Stored procedure examples
'   License:    This code is covered by the Apache license. 
'               For more information read EULA.txt
'--------------------------------------------------------------------
''===========================================================================
#End Region
Imports System.Data.SqlClient
Imports Microsoft.AnalysisServices
Imports Microsoft.AnalysisServices.AdomdServer
Imports System.Diagnostics

Public Class StoredProcedures
    ' ------------------------------------------------------------------------------
    ' Calculates the percentage difference between two decimal values
    ' Call example
    '   CREATE MEMBER CURRENTCUBE.[MEASURES].[PctDiff]
    '   AS AdventureWorks!PctDiff([Measures].[Reseller Sales Amount],
    '        [Measures].[Internet Sales Amount]), 
    '   FORMAT_STRING = "Percent", 
    '   VISIBLE = 1  ;   
    '-------------------------------------------------------------------------------
    Public Shared Function PctDiff(ByVal x As Decimal, ByVal y As Decimal) As Object

        If y = 0 Then
            Return "NA"
        Else
            Return (x - y) / y
        End If

    End Function
    ' ------------------------------------------------------------------------------
    ' Returns the current day
    ' Call example
    ' SELECT {[Internet Sales Amount], [Sales Profit], [Average Profit]} on 0,
    '		StrToMember (AdventureWorks!GetCurrentDay()) on 1
    ' FROM [SOS OLAP]  
    '-------------------------------------------------------------------------------
    Public Shared Function GetCurrentDay() As String
        Return "[Date].[Calendar Date].[2003].[H" + CStr(IIf(Month(Now()) <= 6, 1, 2)) + _
                " CY 2003].[Q" + CStr(DatePart("q", Now())) + " CY 2003].[" + Format(Now(), "MMMM") + "].[" _
                + Format(Now(), "MMMM") + " " + CStr(Day(Now())) + "]"
    End Function
    ' ------------------------------------------------------------------------------
    ' Outputs the fully qualified member names of a set
    ' Call Example
    ' 1. Install DebugView from sysinternals.com and open it.
    ' 2. Use the SQL Server Management Studio to fire an MDX query that returns a set, e.g.
    ' 	SELECT AdventureWorks.ShowSet(
    '	SetToStr( 
    '			{TopCount([Reseller].[Reseller].[Reseller].Members, 5, 
    '			[Reseller Profit Dollar Amount])} 
    '			)
    '            ) ON COLUMNS
    '   FROM [SOS OLAP]	
    '-------------------------------------------------------------------------------
    <CLSCompliant(False)> _
    Public Shared Function ShowSet(ByVal inputSet As String) As [Set]
        System.Diagnostics.Trace.WriteLine(String.Format("Input set: {0}, Pass: {1}", inputSet, Context.Pass.ToString()))
        Return AdomdServer.MDX.StrToSet(inputSet)
    End Function
    ' ------------------------------------------------------------------------------
    ' Filter a set given an expression
    ' Disclaimer -- The C# version of this stored procedure can be found in the AmoAdventureWorks
    '               code sample which is installed with Analysis Services
    '
    ' Call Example
    ' SELECT [Measures].[Internet Sales Amount] ON 0,
    ' AdventureWorks.FilterSet( [Customer].Customer.Members, "NOT IsEmpty([Measures].[Internet Sales Amount])")  on 1
    ' FROM [SOS OLAP]	
    '
    '  or
    ' Select {[Internet Sales Amount], [Sales Profit], [Average Profit]} on 0,
    ' AdventureWorks.FilterSet([Customer].[Customers by Geography].[City].Members,
    ' "[Measures].[Internet Sales Amount] > 250000") on 1
    ' From [SOS OLAP]
    '-------------------------------------------------------------------------------
    <CLSCompliant(False)> _
    Public Shared Function FilterSet(ByVal inputSet As [Set], ByVal filterExpression As String) As [Set]

        Dim tuple As Tuple
        Dim expr As Expression = New Expression(filterExpression)

        Dim resultSetBuilder As SetBuilder = New SetBuilder()

        For Each tuple In inputSet.Tuples
            If CType(expr.Calculate(tuple), Boolean) Then
                resultSetBuilder.Add(tuple)
            End If
        Next

        Return resultSetBuilder.ToSet()
    End Function
    ' ------------------------------------------------------------------------------
    ' Evaluates the values of all tuples in a set and filters those that exceeds the specified ceiling
    ' Call Example
    ' Select {[Internet Sales Amount], [Reseller Sales Amount], [Sales Profit], [Average Profit]} on 0,
    ' AdventureWorks.FilterSetByCeiling(([Customer].[Customers by Geography].[City].Members, 
    ' [Product].[Product by Category].[Subcategory].&[1]), "Measures.[Internet Sales Amount]", 20000) on 1
    ' From [SOS OLAP]
    '-------------------------------------------------------------------------------
    <CLSCompliant(False)> _
    Public Shared Function FilterSetByCeiling(ByVal inputSet As [Set], _
                        ByVal measure As String, ByVal ceiling As Integer) As [Set]

        Dim tuple As Tuple
        Dim expr As Expression = New Expression()
        Dim resultSetBuilder As SetBuilder = New SetBuilder()
        Dim context, expText As String
        Dim mdxValue As MDXValue

        ' Enumerate the tuples in the set
        For Each tuple In inputSet.Tuples
            ' Get the tuple members
            context = GetCurrentContextFromTuple(tuple)
            ' Prepare the expression
            expText = String.Format("({0} {1})", context, measure)
            expr.ExpressionText = expText
            ' Get the value
            mdxValue = expr.Calculate(tuple)

            If mdxValue.ToInt32() <= ceiling Then  ' ! assume the cell value will be int
                resultSetBuilder.Add(tuple)
            End If
        Next

        Return resultSetBuilder.ToSet()
    End Function

    Private Shared Function GetCurrentContextFromTuple(ByVal tuple As Tuple) As String

        Dim context As String = String.Empty
        Dim member As Member

        For Each member In tuple.Members
            context += member.UniqueName + ", "
        Next

        Return context

    End Function
    ' ------------------------------------------------------------------------------
    ' Returns the probability that a given customer may belong to a given cluster
    ' Call Example
    ' SELECT AdventureWorks.GetClusterProbability(40, '10+ miles', 1, 4000, 2, 2, 'Management', 
    '        100000, 'YUP') from [CustomerProfilingMC]
    '-------------------------------------------------------------------------------
    Public Shared Function GetClusterProbability(ByVal age As Integer, _
                                                    ByVal commuteDistance As String, _
                                                    ByVal houseOwner As Integer, _
                                                    ByVal salesAmount As Double, _
                                                    ByVal carsOwned As Integer, _
                                                    ByVal childrenHome As Integer, _
                                                    ByVal occupation As String, _
                                                    ByVal yearlyIncome As Double, _
                                                    ByRef clusterName As String) As Double

        Dim command As New AdomdCommand()

        command.CommandText = "SELECT ClusterProbability('" & clusterName & "') " & _
                                "FROM [" & Context.CurrentMiningModel.Name & "]" & _
                                "NATURAL PREDICTION JOIN " & _
                                "(SELECT @Age as Age, @CommuteDistance as [Commute Distance], " & _
                                "@HouseOwner as [House Owner Flag]," & _
                                "@SalesAmount as [Internet Sales Amount], " & _
                                "@CarsOwned as [Number Cars Owned], " & _
                                "@ChildrenHome as [Number Children At Home]," & _
                                "@Occupation as [Occupation], " & _
                                "@YearlyIncome as [Yearly Income] ) as t "

        command.Parameters.Add("Age", age)
        command.Parameters.Add("CommuteDistance", commuteDistance)
        command.Parameters.Add("HouseOwner", houseOwner)
        command.Parameters.Add("SalesAmount", salesAmount)
        command.Parameters.Add("CarsOwned", carsOwned)
        command.Parameters.Add("ChildrenHome", childrenHome)
        command.Parameters.Add("Occupation", occupation)
        command.Parameters.Add("YearlyIncome", yearlyIncome)

        Dim result As Double = command.ExecuteScalar()

        Return result

    End Function


    ''' <summary>
    ''' Creates a new partition that contains data for the specified number of months
    ''' </summary>
    ''' <param name="cubeName">The Cube Name</param>
    ''' <param name="measureGroupName">The Measure Group where the partition will be created</param>
    ''' <param name="months">The number of months for the partition data</param>
    ''' <remarks> When deploying confiure it for Unrestricted Access
    ''' To call use one of the following to approaches
    '''          1. Use the Management Studio to create an MDX query to SOS OLAP cube and then send the query
    '''              CALL AdventureWorks!CreatePartition("SOS OLAP", "Internet Sales", 3)
    '''          2. Create a SQL Server Analysis Query job with the above statement
    ''' </remarks>
    Public Sub CreatePartition(ByVal cubeName As String, ByVal measureGroupName As String, ByVal months As Integer)

        Dim part As Partition
        Dim mg As MeasureGroup
        Dim maxOrderDateKey As Integer = 0
        Dim svr As New Server()

        svr.Connect("*")

        ' Get the SOS OLAP cube
        Dim db As Database = svr.Databases.GetByName(Context.CurrentDatabaseName)
        Dim cube As Cube = db.Cubes.GetByName(cubeName)

        mg = cube.MeasureGroups.GetByName(measureGroupName)

        ' Get the last partition
        part = mg.Partitions(mg.Partitions.Count - 1)

        If TypeOf part.Source Is QueryBinding = False Then
            Throw New ApplicationException("The partition must have Query Binding")
        End If

        ' Get the last date index from the partition query
        Dim lastDateIndex = GetLastDateIndex(part)

        ' Get the new date index.
        Dim newDateIndex As Integer = GetNewDateIndex(lastDateIndex, months)

        ' If there are no enough data in the partition, exit
        If newDateIndex = 0 Then
            Return
        End If

        ' Get the fact table name from the old partition
        Dim factTableName As String = GetFactTableName(CType(part.Source, QueryBinding).QueryDefinition.ToLower())
        ' Create a new partition
        part = mg.Partitions.Add(measureGroupName & " _" & newDateIndex)
        part.StorageMode = StorageMode.Molap
        ' Create the query text
        Dim queryText As String = String.Format("SELECT * FROM {0} WHERE OrderDateKey >= {1} and OrderDateKey <= {2}", factTableName, lastDateIndex + 1, newDateIndex)
        part.Source = New QueryBinding(db.DataSources(0).ID, queryText)
        ' Save the changes
        part.Update()
        ' Process the partition
        part.Process()

        svr.Disconnect()

    End Sub
    ''' <summary>
    ''' Return the last date key for which data is found in the last partition
    ''' </summary>
    ''' <param name="part">The last partition</param>
    ''' <returns>The last date index assuming it is integer</returns>
    ''' <remarks></remarks>
    Private Function GetLastDateIndex(ByVal part As Partition) As Integer
        ' Get the query definition
        Dim query As String = CType(part.Source, QueryBinding).QueryDefinition
        ' Search for "<=" to get to the last date index, e.g. WHERE OrderDateKey >= 185 AND OrderDateKey <= 549
        Dim pos As Integer = query.IndexOf("<=") + 2
        Dim dateIndex As String = query.Substring(pos).Trim()

        Return Int32.Parse(dateIndex)
    End Function
    ''' <summary>
    ''' Returns the last date index for the new partition scope
    ''' </summary>
    ''' <param name="dateIndex">The last date index</param>
    ''' <param name="months">Month offset</param>
    ''' <returns>The new date index or 0 if no data. </returns>
    ''' <remarks></remarks>
    Private Function GetNewDateIndex(ByVal dateIndex As Integer, ByVal months As Integer) As Integer

        ' Open a connection to the data warehouse
        Dim conn As New SqlConnection("data source=localhost;initial catalog=AdventureWorksDW;Integrated Security=SSPI")

        conn.Open()

        ' Create a command
        Dim cmd As New SqlCommand()
        cmd.Connection = conn

        ' Retrieve the date corresponding to the key 
        cmd.CommandText = "select FullDateAlternateKey from DimTime WHERE TimeKey = " & dateIndex

        Dim tempDate As Date = CDate(cmd.ExecuteScalar())

        ' Add the required number of months 
        tempDate = tempDate.AddMonths(months)

        ' Is the old partition still current?
        If (DateTime.Now < tempDate) Then
            Return 0
        End If

        ' Get the end date index from the Time dimension table
        cmd.CommandText = "select TimeKey from DimTime WHERE FullDateAlternateKey = '" & tempDate.ToString("d") & "'"

        Dim lastIndex As Object = cmd.ExecuteScalar()

        If lastIndex Is Nothing Then
            Return 0
        Else
            Return CInt(lastIndex)
        End If


    End Function
    ''' <summary>
    ''' Returns the fact table name from the partition query
    ''' </summary>
    ''' <param name="query">The partition query</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetFactTableName(ByVal query As String)
        Dim pos As Integer = query.IndexOf("from ") + 5
        Dim endPos As Integer = query.IndexOf("where")
        Return query.Substring(pos, endPos - pos).Trim()
    End Function

    ''' <summary>
    ''' Demonstrates dynamic security
    ''' </summary>
    ''' <param name="userName">User name</param>
    ''' <returns>The regions that the employee is permitted to see</returns>
    ''' <remarks>
    ''' *** USAGE ****
    ''' SELECT [Measures].[Reseller Sales Amount] ON COLUMNS,
    ''' AdventureWorks.GetRegionsByEmployee(username) ON ROWS
    ''' FROM [SOS OLAP]
    ''' </remarks>
    <CLSCompliant(False)> _
    Public Shared Function GetRegionsByEmployee(ByVal userName As String) As [Set]

        ' FOR TESTING PURPOSES ONLY
        ' Strip out the user name from the domain since we will be forcing adventure-works domain
        userName = userName.Substring(userName.IndexOf("\") + 1)

        ' Open a connection to the data warehouse
        Dim conn As New SqlConnection("data source=localhost;initial catalog=AdventureWorksDW;Integrated Security=SSPI")

        conn.Open()

        ' Create a command
        Dim cmd As New SqlCommand()
        cmd.Connection = conn

        ' Retrieve the regions assigned to the employee 
        cmd.CommandText = "SELECT DISTINCT '[Sales Territory].[Region].&[' + CAST(ST.SalesTerritoryKey AS VARCHAR(5)) + ']' AS Region " _
                        & "FROM DimEmployee AS E INNER JOIN " _
                        & "DimSalesTerritory AS ST ON E.SalesTerritoryKey = ST.SalesTerritoryKey " _
                        & "WHERE E.LoginID = 'adventure-works\" & userName & "'"

        cmd.CommandType = CommandType.Text

        ' Get the regions as a reader
        Dim reader As SqlDataReader = cmd.ExecuteReader(CommandBehavior.CloseConnection)

        ' Option 1 - Use StrToSet
        '   Dim stringBuilder As New System.Text.StringBuilder()
        '   Dim regionsString As String
        '   Dim regions As [Set]
        '
        '   While reader.Read()
        '       stringBuilder.Append("," + reader(0).ToString())
        '   End While

        '   regionsString = stringBuilder.ToString()

        '   If regionsString.Trim() <> String.Empty Then
        '       ' Remove last ","
        '       regionsString = regionsString.Remove(0, 1)
        '       ' At this point, we have the set as a string. Now, let's convert it to a set
        '   End If

        '   regionsString = "{" & regionsString & "}"2

        '   regions = MDX.StrToSet(regionsString)
        '   Return regions
        '   End Option 1

        ' Option 2 - Use Expression Builder
        Dim expr As AdomdServer.Expression = New Expression()
        Dim resultSetBuilder As SetBuilder = New SetBuilder()
        Dim tupleBuilder As New TupleBuilder()

        While reader.Read()
            expr.ExpressionText = reader(0).ToString()
            ' Get the member
            Dim member As Member = expr.CalculateMdxObject(Nothing).ToMember()
            ' Construct a tuple
            tupleBuilder.Add(member)
            ' Add the tuple to set
            resultSetBuilder.Add(tupleBuilder.ToTuple())
        End While
        reader.Close()

        Return resultSetBuilder.ToSet()
        ' End Option 2
    End Function
    ''' <summary>
    ''' Determines if a given measure belongs to a given measure group
    ''' </summary>
    ''' <param name="cubeName">The Cube Name</param>
    ''' <param name="measureGroupName">The Measure Group where the partition will be created</param>
    ''' <param name="measureName">The name of the measure sought (Measures.CurrentMember)</param>
    ''' <returns>True if the measure belongs to the measure group. Otherwise, False.</returns>
    ''' <remarks>
    ''' *** USAGE ****
    ''' In the Read Permission field of Cell Data enter
    ''' AdventureWorks.IsMeasureAllowed("SOS OLAP", "Reseller Sales", [Measures].CurrentMember.Name)
    ''' </remarks>
    Public Function IsMeasureAllowed(ByVal cubeName As String, ByVal measureGroupName As String, ByVal measureName As String) As Boolean

        Dim mg As MeasureGroup
        Dim measure As Microsoft.AnalysisServices.Measure
        Dim svr As New Server()
        Dim measureFound As Boolean = False

        svr.Connect("data source=local")

        Dim db As Database = svr.Databases.GetByName(Context.CurrentDatabaseName)
        Dim cube As Cube = db.Cubes.GetByName(cubeName)

        mg = cube.MeasureGroups.GetByName(measureGroupName)

        For Each measure In mg.Measures
            If measure.Name = measureName Then
                measureFound = True
                Exit For
            End If
        Next
        svr.Disconnect()

        Return measureFound

    End Function
End Class

